package provide TR 0.1
package req TM
package req XP
package require nx 
#####################################################################
#
# KaleidoScope - Trace Provider
#
#####################################################################
namespace eval ::TR {
nx::Class create Tracer {
  #one dynamic trace model
  :property -accessor public {traceModel}
  #XMI exporter
  :property -accessor public {exporter}
  #current values of sections and contexts
  :property -accessor public {currentTestBlock "null"}
  :property -accessor public {currentTestScenario "null"}  
  :property -accessor public {currentTestCase "null"}  
  #singleton constructor
  :public object method getInstance {} {
      if {[llength [:info instances]] == 0} {
        set tracer [Tracer new]
        #dynamic tracer has one dynamic trace model
        set tm [::TM::TraceModel getInstance]
        $tm trace set [::TM::Trace new -name trace1]
	$tm testSuite set [::TM::TestSuite new -name ts1]
        $tracer traceModel set $tm 
        return $tracer
        
      } elseif {[llength [:info instances]] == 1} {
        return [:info instances]
      } 
   }
  #check if a component is already registered in trace model 
  # returns the - in any case - registered component	
  :public method checkComponent {co} {
    set ex 0
    foreach c [[:traceModel get] component get] {
      if {[$c name get] eq $co} {
        set ex 1  
        return $c
      } 
    }
    if {!$ex} {
      set newComponent [::TM::Component new -name $co]	    
      set comps [[:traceModel get] component get]
      lappend comps $newComponent
      [:traceModel get] component set $comps    
    }
    return $newComponent
  }


  #check if a class is already registered in trace model 
  # returns the - in any case - registered class
  :public method checkClass {cl} {
    set ex 0
    foreach c [[:traceModel get] class get] {
      if {[$c name get] eq $cl} {
        set ex 1  
        return $c
      } 
    }
    if {!$ex} {
      set newClass [::TM::Class new -name $cl]	    
      set css [[:traceModel get] class get]
      lappend css $newClass
      [:traceModel get] class set $css    
    }
    return $newClass
  }
  #check if an instance is already registered in trace model 
  # returns the - in any case - registered instance
  :public method checkInstance {inst} {
    set exst 0
    foreach i [[:traceModel get] instance get] {
      if {[$i name get] eq $inst} {
        set exst 1  
	return $i     
      } 
    }
    if {!$exst} {
      set newInstance [::TM::Instance new -name $inst]
      set inss [[:traceModel get] instance get]
      lappend inss $newInstance
      [:traceModel get] instance set $inss
    }
    return $newInstance
  }
  #check if a feature is already registered in trace model 
  # returns the - in any case - registered feature 
  :public method checkFeature {feat} {
      set exst 0
      foreach f [[:traceModel get] feature get] {
	  if {[$f name get] eq ""} {			      
	  } elseif {[$f name get] eq $feat} {
	  	set exst 1  
	  	return $f	  	
	} 
      }
      if {!$exst} {
	set newFeature [::TM::Feature new -name $feat]
	set ff [[:traceModel get] feature get]
	lappend ff $newFeature
	[:traceModel get] feature set $ff	
      }
      return $newFeature
    }
    #check if a case is already registered in trace model 
    # returns the in any case registered case	
    :public method checkCase {case} {
	  set exst 0
	  foreach c [[[:traceModel get] testSuite get] case get] {
	    if {$case eq ""} {
		    
	    } elseif {[$c name get] eq [$case name get]} {
	      set case $c
	      set exst 1  
	      break
	    } 
	  }
	  if {!$exst} {
	    set cs [[[:traceModel get] testSuite get] case get]
	    lappend cs $case
	    [[:traceModel get] testSuite get] case set $cs	
	  }
	  return $case
	}
    #check if a scenario is already registered in trace model 
    # returns the in any case registered scenario
    :public method checkScenario {scenario} {
	  set exst 0
	  foreach c [[[:traceModel get] testSuite get] case get] {
		  if {[$c name get] eq [:currentTestCase get]} {
			foreach s [$c scenario get] {
				if {[$s name get] eq ""} {
					    
				} elseif {[$s name get] eq [$scenario name get]} {
					set exst 1  
					set scenario $s 
					break
				} 
			}  
			if {!$exst} {
				set scs [$c scenario get]
				lappend scs $scenario
				$c scenario $scs	
			} 
			return $scenario
			#break
		  }
	  }
	}
  #store current scenario and case
  :public method checkSection {callingObject callingClass} {
    if {$callingClass eq "::STORM::TestScenario"} {
     #:currentTestScenario $callingObject
     :currentTestScenario set [$callingObject name get]
     #check Scenario
    } elseif {$callingClass eq "::STORM::TestCase"} {
     #:currentTestCase $callingObject
     :currentTestCase set [$callingObject name get]
     :currentTestScenario set "null"
     #check case
    } elseif {$callingClass eq "::STORM::TestSuite"} {
     :currentTestCase set "null"
     :currentTestScenario set "null"
    }
  }
  #add a traced call with its dependencies to trace
  :public method addTracedCall {callingMethod callingObject callingClass calledMethod calledObject calledClass arguments returnValue compCalling compCalled} {
		#puts $compCalling 
		#puts $compCalled
		#puts -----------------
		:checkSection $callingObject $callingClass
		set stfCaller "do"
	  #set the feature call
	  set call [::TM::FeatureCall new -name $calledMethod]
	  #set the callee
	  $call callee set [:checkFeature $calledMethod]
	  #set the caller
	  $call caller set [:checkFeature $callingMethod]
	  #set definedBySTF
	  if {[[[$call] caller get] name get] eq $stfCaller} {
	  	[$call] definedBySTF set true 	  
	  }
	  #set the source and its type
	  $call source set [:checkInstance $callingObject]
	  [$call source get] definition set [:checkClass $callingClass]
	  [$call caller get] owningClass set [[$call source get] definition get]
		
		#set the component of the source's type 
		set sourceComponent [:checkComponent $compCalling]
		$sourceComponent addClass [[$call source get] definition get]
		#
		
		#
		#puts [[:traceModel get] component get]
		$sourceComponent addClass [[$call source get] definition get]

	  #set the target and its type
	  $call target set [:checkInstance $calledObject] 
	  [$call target get] definition set [:checkClass $calledClass]
	  [$call callee get] owningClass set [[$call target get] definition get]
		
		#set the component of the target's type
		set targetComponent [:checkComponent $compCalled]
		$targetComponent addClass [[$call target get] definition get]

	  #set the return
	  if {[llength $returnValue]  == 1} {
	    set rt [::TM::ReturnValue new -value $returnValue -name $returnValue]
	    $call returnVal set $rt  
	  } 
	  #set the arguments
	  if {[llength $arguments]  > 0} {
	    foreach arg $arguments {
	      set vs [::TM::Argument new -value $arg -name $arg]
	      $call addArgument $vs
	    }   
	  }
	  #set the block
	  set bl [::TM::TestBlock new -name [:currentTestBlock get]]
	  $call testBlock set $bl
	  #set the scenario
	  set sc [::TM::TestScenario new -name [:currentTestScenario get]]
	  [$call testBlock  get] scenario set $sc
	  #set the case
	  set cs [::TM::TestCase new -name [:currentTestCase get]]
	  #[[$call testBlock] scenario] case $cs
	  #discard programming language internal details
	  if {[[[$call] callee get] name get] in {"init" "new" "__objectparameter"} || 
		  [[[$call] caller get] name get] in {"init" "new" "__objectparameter"} } {
			  
		  #do nothing 
	  #in any other case, add the call
	  } else {
		# assign info that callee 'isConstructor'
	  	if {[[[$call] callee get] name get] eq "create"} {
	  		[[$call] callee get] isConstructor set true
	  		#TODO redefine callee and its type
		}    
		:assignToTestParts $call
		# regularly adding the call to the call sequence
		[[:traceModel get] trace get] addCall $call  
	  }  
  }  
  :public method assignToTestParts {call} {
	  set caseReg 0
	  set caseEx 0
	  #check does not exist
	  if {[:currentTestCase get] eq "null"} {
	  	set caseEx 0
	  #case exists
	  } else {
		set caseEx 1
		foreach c [[[:traceModel get] testSuite get] case get] {
			if {[:currentTestCase get] eq [$c name get]} {
				set currentCase $c
				set caseReg 1
				break
			}	
		}
		if {!$caseReg} {
			set currentCase [::TM::TestCase new -name [:currentTestCase get]]
			set css [[[:traceModel get] testSuite get] case get]
			lappend css $currentCase
			[[:traceModel get] testSuite get] case set $css	
		}
	  }
	  
	  #checkscenario
	  if {$caseEx} {
	  	set scenReg 0
	  	set scenEx 0
		if {[:currentTestScenario get] eq "null"} {
			set scenEx 0	
		} else {
			set scenEx 1
			foreach s [$currentCase scenario get] {
				if {[:currentTestScenario get] eq [$s name get]} {
				set currentScenario $s
				set scenReg 1  
				break
				} 
			}
			if {$scenEx && !$scenReg} {
				set currentScenario [::TM::TestScenario new -name [:currentTestScenario get]]
				set scs [[$currentCase] scenario get]
				lappend scs $currentScenario
				[$currentCase] scenario set $scs	
			}
		} 
	  } 
	  
	  #assignblock to owning test part
	  if {[:currentTestBlock get] eq "null"} {
	  	puts "error in assigning block"	
	  }
	  set blockReg 0
	  if {$caseEx} {
		  #assign block to scenario
		  if {$scenEx} {
			  foreach b [$currentScenario block get] {
				  if {[:currentTestBlock get] eq [$b name get]} {
					  set blockReg 1
					  set currentBlock $b
					  break
				  }
			  }
			  if {!$blockReg} {
				  set currentBlock [::TM::TestBlock new -name [:currentTestBlock get]]
				  set blks [$currentScenario block get]
				  lappend blks $currentBlock
				  $currentScenario block set $blks
			  }			  
  
		  #assign block to case
		  } elseif {!$scenEx} {
			  foreach b [$currentCase block get] {
				  if {[:currentTestBlock get] eq [$b name get]} {
					set blockReg 1	
					set currentBlock $b
					break
				  }
			  }  
			  if {!$blockReg} {
				  set currentBlock [::TM::TestBlock new -name [:currentTestBlock get]]
				  set blks [$currentCase block get]
				  lappend blks $currentBlock
				  $currentCase block set $blks
			  }
		  }
			    	    
	  #assign to testsuite
	  } elseif {!$caseEx} {
		  foreach b [[[:traceModel] testSuite get] block get] {
			  if {[:currentTestBlock get] eq [$b name get]} {
				set blockReg 1	
				set currentBlock $b
				break
			  }
		  }
		  if {!$blockReg} {
			  set currentBlock [::TM::TestBlock new -name [:currentTestBlock get]]
			  set blks [[[:traceModel get] testSuite get] block get]
			  lappend blks $currentBlock
			  [[:traceModel get] testSuite get] block set $blks
		  }		    
	  }

	  #addCallToBlock
	  set callsInBlock [$currentBlock call get]
	  lappend callsInBlock $call
	  $currentBlock call set $callsInBlock
	  
  }
  
  #re-ordering calls according to the principle call-first
  # just handles sut-internal calls of a call depth of 3
  # TODO recursive checking method that handles every call depth...
  
  :public method orderCalls {calls} {
	  set reorderedCalls ""
	  set sCalls1 ""
	  set sCalls2 ""
	  set sCalls3 ""
	  set fC [::TM::FeatureCall new -name "null"]
	  set rootCallerName "do"	  
	  foreach c $calls {
		#start
		if {[$fC name get] eq "null" && [[$c caller get] name get] ne $rootCallerName} {
			puts "error"
		} elseif {[$fC name get] eq "null" && [[$c caller get] name get] eq $rootCallerName} {
			lappend reorderedCalls $c	
		#multiple features called by root caller: already ordered
	  	} elseif {[[$c caller get] name get] eq $rootCallerName && [[$c caller get] name get] eq [[$fC caller get] name get]} {
	  		lappend reorderedCalls $c	
		#feature call chain triggered by root caller			    
		} elseif {[[$c callee get] name get] eq [[$fC caller get] name get] && [[$c caller get] name get] eq $rootCallerName} {
			lappend reorderedCalls $c
			
			foreach s1 $sCalls1 {
				if {[[$s1 caller get] name get] eq [[$c callee get] name get] } { 
					lappend reorderedCalls $s1
					
					foreach s2 $sCalls2 {
						if {[[$s2 caller get] name get] eq [[$s1 callee get] name get] } { 
							lappend reorderedCalls $s2
							
							foreach s3 $sCalls3 {
								if {[[$s3 caller get] name get] eq [[$s2 callee get] name get] } {
									lappend reorderedCalls $s3
								}	
							}
							set sCalls3 ""
							
						} else {
							lappend sCalls3 $s2
						} 
					}
					set sCalls2 ""
				} else {
					lappend sCalls2 $s1
				}
			}	
			set sCalls1 ""
		  #no order  
		} elseif {[[$c caller get] name get] ne [[$fC caller get] name get] || [[$c caller get] name get] ne [[$fC callee get] name get]} {
			lappend sCalls1 $c
		}    
		set fC $c
	  }
	  [[:traceModel get] trace get] calls set $reorderedCalls
	 
	  return $reorderedCalls
  }
  
} 

  
}  
