package require nx 
package provide XP 0.1
#####################################################################
#
# KaleidoScope - Trace-XMI Exporter
#
#####################################################################
namespace eval ::XP {
  nx::Class create XMIExporter { 
    :property -accessor public {traceModel}  
    :public method createXMITraceModel { } {
      set xmiText [:buildHeader]
      append xmiText [:buildTraceModel]
      set xmiText [string map {:: "" # "" } $xmiText]
      set path [file join [pwd] [[:traceModel get] name get].xmi]
      set f $path
      set fileId [open $f "w"]
      puts -nonewline $fileId $xmiText
      close $fileId
      puts "XMI trace model **[[:traceModel get] name get]** can be found under $path"
    }
    :public method buildHeader {} {
      return "<?xml version='1.0' encoding='ASCII'?>"
    }
    :public method buildTraceModel { } {
      set traceModelTag "<TM:TraceModel xmi:version='2.0' xmlns:xmi='http://www.omg.org/XMI' xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' xmlns:TM='traceModel' xsi:schemaLocation='traceModel EXTTRACE.ecore' name='tm1'>"
      foreach f [[:traceModel get] feature get] {
	append traceModelTag [:buildFeature $f]
      }
      foreach c [[:traceModel get] class get] {
	append traceModelTag [:buildClass $c]
      }
      foreach i [[:traceModel get] instance get] {
	append traceModelTag [:buildInstance $i]
      }
#puts test!!![[:traceModel get] component get]
			foreach co [[:traceModel get] component get] {
	append traceModelTag [:buildComponent $co]
      }
      append traceModelTag [:buildSuite [[:traceModel get] testSuite get]]
      append traceModelTag [:buildTrace [[:traceModel get] trace get]] 
      append traceModelTag " </TM:TraceModel>"
      return $traceModelTag
    }	
    :public method buildFeature {feature} {
      set featureTag " <feature name='[$feature name get]' isConstructor='[$feature isConstructor get]' isDestructor='[$feature isDestructor get]' owningClass='//@class.[lsearch [[:traceModel get] class get] [$feature owningClass get]]'/>"
      #puts xf-owningClass[[$feature owningClass] name]
      return $featureTag
    }
    :public method buildClass {class} {
      set classTag " <class name='[$class name get]'"
      set first 1
      foreach oF [$class ownedFeature get] {
	if $first {
	  set first 0
	  append classTag " ownedFeature="
	}
	append classTag " //@feature.0"
	
      }		
      append classTag "/>"	
      return $classTag
    }
		##
		:public method buildComponent {component} {
			set componentTag "<component name='[$component name get]'"
			set first 1
			foreach cl [$component class get] {
				if $first {
					set first 0
					append componentTag " class='"
				}
				append componentTag "//@class.[lsearch [[:traceModel get] class get] $cl] "
				#puts yy[lsearch [[:traceModel get] class get] $cl]
				#puts xx[$cl name get]
			}
			if !$first {
				append componentTag "'"
			}
			append componentTag "/>"
			return $componentTag
		}
		##


    :public method buildInstance {instance} {
      set instanceTag " <instance name='[$instance name get]' definition='//@class.[lsearch [[:traceModel get] class get] [$instance definition get]]'/>"
      return $instanceTag
    }
    :public method buildSuite {suite} {
      set suiteTag "<testSuite name='[$suite name get]'>"
      foreach b [$suite block get] {
	append suiteTag [:buildBlock $b]
      }
      foreach c [$suite case get] {
	append suiteTag [:buildCase $c]
      }
      append suiteTag "</testSuite>" 
      return $suiteTag
    }
    :public method buildCase {case} {
      set caseTag "<case name='[$case name get]'>"
      foreach b [$case block get] {
	append caseTag [:buildBlock $b]
      }
      foreach s [$case scenario get] { 
	append caseTag [:buildScenario $s]
      }
      append caseTag "</case>"
      return $caseTag
    }	
    :public method buildScenario {scenario} {
      set scenarioTag "<scenario name='[$scenario name get]'>"
      foreach b [$scenario block get] {
	append scenarioTag [:buildBlock $b]		
      }
      append scenarioTag "</scenario>"
      return $scenarioTag
    }	
    :public method buildBlock {block} {
      set blockTag "<block name='[$block name get]'"
      set first 1
      foreach c [$block call get] {
	if $first {
	  append blockTag " call='"
	  set first 0
	}
	append blockTag "//@trace/@call.[lsearch [[[:traceModel get] trace get] calls get] $c] "
      }
      append blockTag "'/>"
      return $blockTag
    }
    :public method buildCall {call} {
      set callTag "<call source='//@instance.[lsearch [[:traceModel get] instance get] [$call source get]]' target='//@instance.[lsearch [[:traceModel get] instance get] [$call target get]]' caller='//@feature.[lsearch [[:traceModel get] feature get] [$call caller get]]' callee='//@feature.[lsearch [[:traceModel get] feature get] [$call callee get]]' name='[$call name get]' definedBySTF='[$call definedBySTF get]'>"
      foreach a [$call argument get] {
	append callTag "<argument name='[$a name get]'/>"
      }
      foreach rv [$call returnVal get] {
	append callTag "<returnValue name='[$rv name get]'/>"
      }
      append callTag "</call>"
      return $callTag
    }
    :public method buildTrace {trace} {
      set traceTag "<trace name='t1'>"
      foreach c [$trace calls get] {
	append traceTag [:buildCall $c]
      }
      append traceTag "</trace>"
      return $traceTag
    } 
  }	 
}
